@@ -4,7 +4,7 @@
  *
  * Copyright (C) 1993 Theodore Ts'o.  This file may be redistributed
  * under the terms of the GNU Public License.
- * 
+ *
  * Modifications by Robert Sanders <gt8134b@prism.gatech.edu>
  */
 
@@ -16,7 +16,7 @@
 #include <time.h>
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#else 
+#else
 extern int optind;
 extern char *optarg;
 #endif
@@ -36,6 +36,7 @@ extern char *optarg;
 #include <ext2fs/ext2_ext_attr.h>
 
 #include "../version.h"
+#include "jfs_user.h"
 
 extern ss_request_table debug_cmds;
 ss_request_table *extra_cmds;
@@ -45,7 +46,7 @@ ext2_filsys	current_fs = NULL;
 ext2_ino_t	root, cwd;
 
 static void open_filesystem(char *device, int open_flags, blk_t superblock,
-			    blk_t blocksize, int catastrophic, 
+			    blk_t blocksize, int catastrophic,
 			    char *data_filename)
 {
 	int	retval;
@@ -59,7 +60,7 @@ static void open_filesystem(char *device, int open_flags, blk_t superblock,
 
 	if (data_filename) {
 		if ((open_flags & EXT2_FLAG_IMAGE_FILE) == 0) {
-			com_err(device, 0, 
+			com_err(device, 0,
 				"The -d option is only valid when reading an e2image file");
 			current_fs = NULL;
 			return;
@@ -77,7 +78,7 @@ static void open_filesystem(char *device, int open_flags, blk_t superblock,
 			"opening read-only because of catastrophic mode");
 		open_flags &= ~EXT2_FLAG_RW;
 	}
-	
+
 	retval = ext2fs_open(device, open_flags, superblock, blocksize,
 			     unix_io_manager, &current_fs);
 	if (retval) {
@@ -104,7 +105,7 @@ static void open_filesystem(char *device, int open_flags, blk_t superblock,
 	if (data_io) {
 		retval = ext2fs_set_data_io(current_fs, data_io);
 		if (retval) {
-			com_err(device, retval, 
+			com_err(device, retval,
 				"while setting data source");
 			goto errout;
 		}
@@ -128,9 +129,9 @@ void do_open_filesys(int argc, char **argv)
 	blk_t	blocksize = 0;
 	int	open_flags = EXT2_FLAG_SOFTSUPP_FEATURES;
 	char	*data_filename = 0;
-	
+
 	reset_getopt();
-	while ((c = getopt (argc, argv, "iwfecb:s:d:")) != EOF) {
+	while ((c = getopt (argc, argv, "iwfecb:s:d:D")) != EOF) {
 		switch (c) {
 		case 'i':
 			open_flags |= EXT2_FLAG_IMAGE_FILE;
@@ -150,6 +151,9 @@ void do_open_filesys(int argc, char **argv)
 		case 'd':
 			data_filename = optarg;
 			break;
+		case 'D':
+			open_flags |= EXT2_FLAG_DIRECT_IO;
+			break;
 		case 'b':
 			blocksize = parse_ulong(optarg, argv[0],
 						"block size", &err);
@@ -172,7 +176,7 @@ void do_open_filesys(int argc, char **argv)
 	if (check_fs_not_open(argv[0]))
 		return;
 	open_filesystem(argv[optind], open_flags,
-			superblock, blocksize, catastrophic, 
+			superblock, blocksize, catastrophic,
 			data_filename);
 	return;
 
@@ -199,7 +203,7 @@ void do_lcd(int argc, char **argv)
 static void close_filesystem(NOARGS)
 {
 	int	retval;
-	
+
 	if (current_fs->flags & EXT2_FLAG_IB_DIRTY) {
 		retval = ext2fs_write_inode_bitmap(current_fs);
 		if (retval)
@@ -219,8 +223,28 @@ static void close_filesystem(NOARGS)
 
 void do_close_filesys(int argc, char **argv)
 {
-	if (common_args_process(argc, argv, 1, 1, "close_filesys", "", 0))
+	int	c;
+
+	if (check_fs_open(argv[0]))
 		return;
+
+	reset_getopt();
+	while ((c = getopt (argc, argv, "a")) != EOF) {
+		switch (c) {
+		case 'a':
+			current_fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;
+			break;
+		default:
+			goto print_usage;
+		}
+	}
+
+	if (argc > optind) {
+	print_usage:
+		com_err(0, 0, "Usage: close_filesys [-a]");
+		return;
+	}
+
 	close_filesystem();
 }
 
@@ -229,7 +253,7 @@ void do_init_filesys(int argc, char **argv)
 	struct ext2_super_block param;
 	errcode_t	retval;
 	int		err;
-	
+
 	if (common_args_process(argc, argv, 3, 3, "initialize",
 				"<device> <blocksize>", CHECK_FS_NOTOPEN))
 		return;
@@ -290,9 +314,6 @@ void do_show_super_stats(int argc, char *argv[])
 	int	c, header_only = 0;
 	int	numdirs = 0, first, gdt_csum;
 
-	gdt_csum = EXT2_HAS_RO_COMPAT_FEATURE(current_fs->super,
-					      EXT4_FEATURE_RO_COMPAT_GDT_CSUM);
-
 	reset_getopt();
 	while ((c = getopt (argc, argv, "h")) != EOF) {
 		switch (c) {
@@ -314,12 +335,14 @@ void do_show_super_stats(int argc, char *argv[])
 	for (i=0; i < current_fs->group_desc_count; i++)
 		numdirs += current_fs->group_desc[i].bg_used_dirs_count;
 	fprintf(out, "Directories:              %d\n", numdirs);
-	
+
 	if (header_only) {
 		close_pager(out);
 		return;
 	}
-	
+
+	gdt_csum = EXT2_HAS_RO_COMPAT_FEATURE(current_fs->super,
+					      EXT4_FEATURE_RO_COMPAT_GDT_CSUM);
 	gdp = &current_fs->group_desc[0];
 	for (i = 0; i < current_fs->group_desc_count; i++, gdp++) {
 		fprintf(out, " Group %2d: block bitmap at %u, "
@@ -360,7 +383,7 @@ print_usage:
 	fprintf(stderr, "%s: Usage: show_super [-h]\n", argv[0]);
 }
 
-void do_dirty_filesys(int argc EXT2FS_ATTR((unused)), 
+void do_dirty_filesys(int argc EXT2FS_ATTR((unused)),
 		      char **argv EXT2FS_ATTR((unused)))
 {
 	if (check_fs_open(argv[0]))
@@ -401,10 +424,10 @@ static void finish_range(struct list_blocks_struct *lb)
 	lb->first_block = 0;
 }
 
-static int list_blocks_proc(ext2_filsys fs EXT2FS_ATTR((unused)), 
-			    blk_t *blocknr, e2_blkcnt_t blockcnt, 
+static int list_blocks_proc(ext2_filsys fs EXT2FS_ATTR((unused)),
+			    blk_t *blocknr, e2_blkcnt_t blockcnt,
 			    blk_t ref_block EXT2FS_ATTR((unused)),
-			    int ref_offset EXT2FS_ATTR((unused)), 
+			    int ref_offset EXT2FS_ATTR((unused)),
 			    void *private)
 {
 	struct list_blocks_struct *lb = (struct list_blocks_struct *) private;
@@ -450,7 +473,7 @@ static void dump_xattr_string(FILE *out, const char *str, int len)
 {
 	int printable = 0;
 	int i;
-	
+
 	/* check: is string "printable enough?" */
 	for (i = 0; i < len; i++)
 		if (isprint(str[i]))
@@ -467,9 +490,9 @@ static void dump_xattr_string(FILE *out, const char *str, int len)
 			fprintf(out, "%02x ", (unsigned char)str[i]);
 }
 
-static void internal_dump_inode_extra(FILE *out, 
+static void internal_dump_inode_extra(FILE *out,
 				      const char *prefix EXT2FS_ATTR((unused)),
-				      ext2_ino_t inode_num EXT2FS_ATTR((unused)), 
+				      ext2_ino_t inode_num EXT2FS_ATTR((unused)),
 				      struct ext2_inode_large *inode)
 {
 	struct ext2_ext_attr_entry *entry;
@@ -503,7 +526,7 @@ static void internal_dump_inode_extra(FILE *out,
 				return;
 			}
 			fprintf(out, "  ");
-			dump_xattr_string(out, EXT2_EXT_ATTR_NAME(entry), 
+			dump_xattr_string(out, EXT2_EXT_ATTR_NAME(entry),
 					  entry->e_name_len);
 			fprintf(out, " = \"");
 			dump_xattr_string(out, start + entry->e_value_offs,
@@ -531,6 +554,136 @@ static void dump_blocks(FILE *f, const char *prefix, ext2_ino_t inode)
 	fprintf(f,"\n");
 }
 
+static int int_log10(unsigned long long arg)
+{
+	int     l = 0;
+
+	arg = arg / 10;
+	while (arg) {
+		l++;
+		arg = arg / 10;
+	}
+	return l;
+}
+
+#define DUMP_LEAF_EXTENTS	0x01
+#define DUMP_NODE_EXTENTS	0x02
+#define DUMP_EXTENT_TABLE	0x04
+
+static void dump_extents(FILE *f, const char *prefix, ext2_ino_t ino,
+			 int flags, int logical_width, int physical_width)
+{
+	ext2_extent_handle_t	handle;
+	struct ext2fs_extent	extent;
+	struct ext2_extent_info info;
+	int			op = EXT2_EXTENT_ROOT;
+	unsigned int		printed = 0;
+	errcode_t 		errcode;
+
+	errcode = ext2fs_extent_open(current_fs, ino, &handle);
+	if (errcode)
+		return;
+
+	if (flags & DUMP_EXTENT_TABLE)
+		fprintf(f, "Level Entries %*s %*s Length Flags\n",
+			(logical_width*2)+3, "Logical",
+			(physical_width*2)+3, "Physical");
+	else
+		fprintf(f, "%sEXTENTS:\n%s", prefix, prefix);
+
+	while (1) {
+		errcode = ext2fs_extent_get(handle, op, &extent);
+
+		if (errcode)
+			break;
+
+		op = EXT2_EXTENT_NEXT;
+
+		if (extent.e_flags & EXT2_EXTENT_FLAGS_SECOND_VISIT)
+			continue;
+
+		if (extent.e_flags & EXT2_EXTENT_FLAGS_LEAF) {
+			if ((flags & DUMP_LEAF_EXTENTS) == 0)
+				continue;
+		} else {
+			if ((flags & DUMP_NODE_EXTENTS) == 0)
+				continue;
+		}
+
+		errcode = ext2fs_extent_get_info(handle, &info);
+		if (errcode)
+			continue;
+
+		if (!(extent.e_flags & EXT2_EXTENT_FLAGS_LEAF)) {
+			if (extent.e_flags & EXT2_EXTENT_FLAGS_SECOND_VISIT)
+				continue;
+
+			if (flags & DUMP_EXTENT_TABLE) {
+				fprintf(f, "%2d/%2d %3d/%3d %*llu - %*llu "
+					"%*llu%*s %6u\n",
+					info.curr_level, info.max_depth,
+					info.curr_entry, info.num_entries,
+					logical_width,
+					extent.e_lblk,
+					logical_width,
+					extent.e_lblk + (extent.e_len - 1),
+					physical_width,
+					extent.e_pblk,
+					physical_width+3, "", extent.e_len);
+				continue;
+			}
+
+			fprintf(f, "%s(ETB%d):%lld",
+				printed ? ", " : "", info.curr_level,
+				extent.e_pblk);
+			printed = 1;
+			continue;
+		}
+
+		if (flags & DUMP_EXTENT_TABLE) {
+			fprintf(f, "%2d/%2d %3d/%3d %*llu - %*llu "
+				"%*llu - %*llu %6u %s\n",
+				info.curr_level, info.max_depth,
+				info.curr_entry, info.num_entries,
+				logical_width,
+				extent.e_lblk,
+				logical_width,
+				extent.e_lblk + (extent.e_len - 1),
+				physical_width,
+				extent.e_pblk,
+				physical_width,
+				extent.e_pblk + (extent.e_len - 1),
+				extent.e_len,
+				extent.e_flags & EXT2_EXTENT_FLAGS_UNINIT ?
+					"Uninit" : "");
+			continue;
+		}
+
+		if (extent.e_len == 0)
+			continue;
+		else if (extent.e_len == 1)
+			fprintf(f,
+				"%s(%lld%s):%lld",
+				printed ? ", " : "",
+				extent.e_lblk,
+				extent.e_flags & EXT2_EXTENT_FLAGS_UNINIT ?
+				"[u]" : "",
+				extent.e_pblk);
+		else
+			fprintf(f,
+				"%s(%lld-%lld%s):%lld-%lld",
+				printed ? ", " : "",
+				extent.e_lblk,
+				extent.e_lblk + (extent.e_len - 1),
+				extent.e_flags & EXT2_EXTENT_FLAGS_UNINIT ?
+					"[u]" : "",
+				extent.e_pblk,
+				extent.e_pblk + (extent.e_len - 1));
+		printed = 1;
+	}
+	if (printed)
+		fprintf(f, "\n");
+}
 
 void internal_dump_inode(FILE *out, const char *prefix,
 			 ext2_ino_t inode_num, struct ext2_inode *inode,
@@ -558,11 +711,11 @@ void internal_dump_inode(FILE *out, const char *prefix,
 	fprintf(out, "%sMode:  %04o   Flags: 0x%x\n",
 		prefix, inode->i_mode & 0777, inode->i_flags);
 	if (is_large_inode && large_inode->i_extra_isize >= 24) {
-		fprintf(out, "%sGeneration: %u    Version: 0x%08x:%08x\n", 
+		fprintf(out, "%sGeneration: %u    Version: 0x%08x:%08x\n",
 			prefix, inode->i_generation, large_inode->i_version_hi,
 			inode->osd1.linux1.l_i_version);
 	} else {
-		fprintf(out, "%sGeneration: %u    Version: 0x%08x\n", prefix, 
+		fprintf(out, "%sGeneration: %u    Version: 0x%08x\n", prefix,
 			inode->i_generation, inode->osd1.linux1.l_i_version);
 	}
 	fprintf(out, "%sUser: %5d   Group: %5d   Size: ",
@@ -581,14 +734,16 @@ void internal_dump_inode(FILE *out, const char *prefix,
 			inode->i_file_acl, LINUX_S_ISDIR(inode->i_mode) ? inode->i_dir_acl : 0,
 			inode->osd1.hurd1.h_i_translator);
 	else
-		fprintf(out, "%sFile ACL: %d    Directory ACL: %d\n",
+		fprintf(out, "%sFile ACL: %llu    Directory ACL: %d\n",
 			prefix,
-			inode->i_file_acl, LINUX_S_ISDIR(inode->i_mode) ? inode->i_dir_acl : 0);
-	if (os == EXT2_OS_LINUX) 
+			inode->i_file_acl | ((long long)
+				(inode->osd2.linux2.l_i_file_acl_high) << 32),
+			LINUX_S_ISDIR(inode->i_mode) ? inode->i_dir_acl : 0);
+	if (os == EXT2_OS_LINUX)
 		fprintf(out, "%sLinks: %d   Blockcount: %llu\n",
-			prefix, inode->i_links_count, 
-			(((unsigned long long) 
-			  inode->osd2.linux2.l_i_blocks_hi << 32)) + 
+			prefix, inode->i_links_count,
+			(((unsigned long long)
+			  inode->osd2.linux2.l_i_blocks_hi << 32)) +
 			inode->i_blocks);
 	else
 		fprintf(out, "%sLinks: %d   Blockcount: %u\n",
@@ -604,10 +759,10 @@ void internal_dump_inode(FILE *out, const char *prefix,
 	fprintf(out, "%sFragment:  Address: %d    Number: %d    Size: %d\n",
 		prefix, inode->i_faddr, frag, fsize);
 	if (is_large_inode && large_inode->i_extra_isize >= 24) {
-		fprintf(out, "%s ctime: 0x%08x:%08x -- %s", prefix, 
+		fprintf(out, "%s ctime: 0x%08x:%08x -- %s", prefix,
 			inode->i_ctime, large_inode->i_ctime_extra,
 			time_to_string(inode->i_ctime));
-		fprintf(out, "%s atime: 0x%08x:%08x -- %s", prefix, 
+		fprintf(out, "%s atime: 0x%08x:%08x -- %s", prefix,
 			inode->i_atime, large_inode->i_atime_extra,
 			time_to_string(inode->i_atime));
 		fprintf(out, "%s mtime: 0x%08x:%08x -- %s", prefix,
@@ -624,7 +779,7 @@ void internal_dump_inode(FILE *out, const char *prefix,
 		fprintf(out, "%smtime: 0x%08x -- %s", prefix, inode->i_mtime,
 			time_to_string(inode->i_mtime));
 	}
-	if (inode->i_dtime) 
+	if (inode->i_dtime)
 	  fprintf(out, "%sdtime: 0x%08x -- %s", prefix, inode->i_dtime,
 		  time_to_string(inode->i_dtime));
 	if (EXT2_INODE_SIZE(current_fs->super) > EXT2_GOOD_OLD_INODE_SIZE)
@@ -643,21 +798,25 @@ void internal_dump_inode(FILE *out, const char *prefix,
 			devnote = "";
 		} else {
 			major = (inode->i_block[1] & 0xfff00) >> 8;
-			minor = ((inode->i_block[1] & 0xff) | 
+			minor = ((inode->i_block[1] & 0xff) |
 				 ((inode->i_block[1] >> 12) & 0xfff00));
 			devnote = "(New-style) ";
 		}
-		fprintf(out, "%sDevice major/minor number: %02d:%02d (hex %02x:%02x)\n", 
+		fprintf(out, "%sDevice major/minor number: %02d:%02d (hex %02x:%02x)\n",
 			devnote, major, minor, major, minor);
+	} else if (do_dump_blocks) {
+		if (inode->i_flags & EXT4_EXTENTS_FL)
+			dump_extents(out, prefix, inode_num,
+				     DUMP_LEAF_EXTENTS|DUMP_NODE_EXTENTS, 0, 0);
+		else
+			dump_blocks(out, prefix, inode_num);
 	}
-	else if (do_dump_blocks)
-		dump_blocks(out, prefix, inode_num);
 }
 
 static void dump_inode(ext2_ino_t inode_num, struct ext2_inode *inode)
 {
 	FILE	*out;
-	
+
 	out = open_pager();
 	internal_dump_inode(out, "", inode_num, inode, 1);
 	close_pager(out);
@@ -694,6 +853,69 @@ void do_stat(int argc, char *argv[])
 	return;
 }
 
+void do_dump_extents(int argc, char *argv[])
+{
+	struct ext2_inode inode;
+	ext2_ino_t	ino;
+	FILE		*out;
+	int		c, flags = 0;
+	int		logical_width;
+	int		physical_width;
+
+	reset_getopt();
+	while ((c = getopt(argc, argv, "nl")) != EOF) {
+		switch (c) {
+		case 'n':
+			flags |= DUMP_NODE_EXTENTS;
+			break;
+		case 'l':
+			flags |= DUMP_LEAF_EXTENTS;
+			break;
+		}
+	}
+
+	if (argc != optind+1) {
+	print_usage:
+		com_err(0, 0, "Usage: dump_extents [-n] [-l] file");
+		return;
+	}
+
+	if (flags == 0)
+		flags = DUMP_NODE_EXTENTS | DUMP_LEAF_EXTENTS;
+	flags |= DUMP_EXTENT_TABLE;
+
+	if (check_fs_open(argv[0]))
+		return;
+
+	ino = string_to_inode(argv[optind]);
+	if (ino == 0)
+		return;
+
+	if (debugfs_read_inode(ino, &inode, argv[0]))
+		return;
+
+	if ((inode.i_flags & EXT4_EXTENTS_FL) == 0) {
+		fprintf(stderr, "%s: does not uses extent block maps\n",
+			argv[optind]);
+		return;
+	}
+
+	logical_width = int_log10(((inode.i_size |
+				    (__u64) inode.i_size_high << 32) +
+				   current_fs->blocksize - 1) /
+				  current_fs->blocksize) + 1;
+	if (logical_width < 5)
+		logical_width = 5;
+	physical_width = int_log10(current_fs->super->s_blocks_count) + 1;
+	if (physical_width < 5)
+		physical_width = 5;
+
+	out = open_pager();
+	dump_extents(out, "", ino, flags, logical_width, physical_width);
+	close_pager(out);
+	return;
+}
+
 void do_chroot(int argc, char *argv[])
 {
 	ext2_ino_t inode;
@@ -899,7 +1121,7 @@ void do_modify_inode(int argc, char *argv[])
 	const char	*octal_format = "0%o";
 	const char	*decimal_format = "%d";
 	const char	*unsignedlong_format = "%lu";
-	
+
 	if (common_inode_args_process(argc, argv, &inode_num, CHECK_FS_RW))
 		return;
 
@@ -907,7 +1129,7 @@ void do_modify_inode(int argc, char *argv[])
 
 	if (debugfs_read_inode(inode_num, &inode, argv[1]))
 		return;
-	
+
 	modify_u16(argv[0], "Mode", octal_format, &inode.i_mode);
 	modify_u16(argv[0], "User ID", decimal_format, &inode.i_uid);
 	modify_u16(argv[0], "Group ID", decimal_format, &inode.i_gid);
@@ -918,7 +1140,7 @@ void do_modify_inode(int argc, char *argv[])
 	modify_u32(argv[0], "Deletion time", decimal_format, &inode.i_dtime);
 	modify_u16(argv[0], "Link count", decimal_format, &inode.i_links_count);
 	if (os == EXT2_OS_LINUX)
-		modify_u16(argv[0], "Block count high", unsignedlong_format, 
+		modify_u16(argv[0], "Block count high", unsignedlong_format,
 			   &inode.osd2.linux2.l_i_blocks_hi);
 	modify_u32(argv[0], "Block count", unsignedlong_format, &inode.i_blocks);
 	modify_u32(argv[0], "File flags", hex_format, &inode.i_flags);
@@ -935,7 +1157,7 @@ void do_modify_inode(int argc, char *argv[])
 	if (os == EXT2_OS_HURD)
 		modify_u32(argv[0], "Translator Block",
 			    decimal_format, &inode.osd1.hurd1.h_i_translator);
-	
+
 	modify_u32(argv[0], "Fragment address", decimal_format, &inode.i_faddr);
 	switch (os) {
 	    case EXT2_OS_HURD:
@@ -955,7 +1177,7 @@ void do_modify_inode(int argc, char *argv[])
 		modify_u32(argv[0], buf, decimal_format, &inode.i_block[i]);
 	}
 	modify_u32(argv[0], "Indirect Block", decimal_format,
-		    &inode.i_block[EXT2_IND_BLOCK]);    
+		    &inode.i_block[EXT2_IND_BLOCK]);
 	modify_u32(argv[0], "Double Indirect Block", decimal_format,
 		    &inode.i_block[EXT2_DIND_BLOCK]);
 	modify_u32(argv[0], "Triple Indirect Block", decimal_format,
@@ -968,7 +1190,7 @@ void do_change_working_dir(int argc, char *argv[])
 {
 	ext2_ino_t	inode;
 	int		retval;
-	
+
 	if (common_inode_args_process(argc, argv, &inode, 0))
 		return;
 
@@ -985,7 +1207,7 @@ void do_print_working_directory(int argc, char *argv[])
 {
 	int	retval;
 	char	*pathname = NULL;
-	
+
 	if (common_args_process(argc, argv, 1, 1,
 				"print_working_directory", "", 0))
 		return;
@@ -1025,22 +1247,22 @@ static int ext2_file_type(unsigned int mode)
 
 	if (LINUX_S_ISDIR(mode))
 		return EXT2_FT_DIR;
-	
+
 	if (LINUX_S_ISCHR(mode))
 		return EXT2_FT_CHRDEV;
-	
+
 	if (LINUX_S_ISBLK(mode))
 		return EXT2_FT_BLKDEV;
-	
+
 	if (LINUX_S_ISLNK(mode))
 		return EXT2_FT_SYMLINK;
 
 	if (LINUX_S_ISFIFO(mode))
 		return EXT2_FT_FIFO;
-	
+
 	if (LINUX_S_ISSOCK(mode))
 		return EXT2_FT_SOCK;
-	
+
 	return 0;
 }
 
@@ -1065,7 +1287,7 @@ static void make_link(char *sourcename, char *destname)
 		base_name = sourcename;
 	/*
 	 * Figure out the destination.  First see if it exists and is
-	 * a directory.  
+	 * a directory.
 	 */
 	if (! (retval=ext2fs_namei(current_fs, root, cwd, destname, &dir)))
 		dest = base_name;
@@ -1089,8 +1311,8 @@ static void make_link(char *sourcename, char *destname)
 
 	if (debugfs_read_inode(ino, &inode, sourcename))
 		return;
-	
-	retval = ext2fs_link(current_fs, dir, dest, ino, 
+
+	retval = ext2fs_link(current_fs, dir, dest, ino,
 			     ext2_file_type(inode.i_mode));
 	if (retval)
 		com_err("make_link", retval, 0);
@@ -1108,7 +1330,7 @@ void do_link(int argc, char *argv[])
 }
 
 static int mark_blocks_proc(ext2_filsys fs, blk_t *blocknr,
-			    int blockcnt EXT2FS_ATTR((unused)), 
+			    int blockcnt EXT2FS_ATTR((unused)),
 			    void *private EXT2FS_ATTR((unused)))
 {
 	blk_t	block;
@@ -1142,7 +1364,7 @@ void do_undel(int argc, char *argv[])
 
 	/*
 	 * XXX this function doesn't handle changing the links count on the
-	 * parent directory when undeleting a directory.  
+	 * parent directory when undeleting a directory.
 	 */
 	inode.i_links_count = LINUX_S_ISDIR(inode.i_mode) ? 2 : 1;
 	inode.i_dtime = 0;
@@ -1164,7 +1386,7 @@ static void unlink_file_by_name(char *filename)
 	int		retval;
 	ext2_ino_t	dir;
 	char		*base_name;
-	
+
 	base_name = strrchr(filename, '/');
 	if (base_name) {
 		*base_name++ = '\0';
@@ -1196,7 +1418,7 @@ void do_find_free_block(int argc, char *argv[])
  	int		count;
 	errcode_t	retval;
 	char		*tmp;
-	
+
 	if ((argc > 3) || (argc==2 && *argv[1] == '?')) {
 		com_err(argv[0], 0, "Usage: find_free_block [count [goal]]");
 		return;
@@ -1224,7 +1446,7 @@ void do_find_free_block(int argc, char *argv[])
 		goal = current_fs->super->s_first_data_block;
 
 	printf("Free blocks found: ");
-	free_blk = goal - 1;	
+	free_blk = goal - 1;
 	while (count-- > 0) {
 		retval = ext2fs_new_block(current_fs, free_blk + 1, 0,
 					  &free_blk);
@@ -1248,7 +1470,7 @@ void do_find_free_inode(int argc, char *argv[])
 	int		mode;
 	int		retval;
 	char		*tmp;
-	
+
 	if (argc > 3 || (argc>1 && *argv[1] == '?')) {
 		com_err(argv[0], 0, "Usage: find_free_inode [dir] [mode]");
 		return;
@@ -1366,6 +1588,7 @@ void do_write(int argc, char *argv[])
 		retval = ext2fs_expand_dir(current_fs, cwd);
 		if (retval) {
 			com_err(argv[0], retval, "while expanding directory");
+			close(fd);
 			return;
 		}
 		retval = ext2fs_link(current_fs, cwd, argv[2], newfile,
@@ -1381,10 +1604,13 @@ void do_write(int argc, char *argv[])
 	ext2fs_inode_alloc_stats2(current_fs, newfile, +1, 0);
 	memset(&inode, 0, sizeof(inode));
 	inode.i_mode = (statbuf.st_mode & ~LINUX_S_IFMT) | LINUX_S_IFREG;
-	inode.i_atime = inode.i_ctime = inode.i_mtime = 
+	inode.i_atime = inode.i_ctime = inode.i_mtime =
 		current_fs->now ? current_fs->now : time(0);
 	inode.i_links_count = 1;
 	inode.i_size = statbuf.st_size;
+	if (current_fs->super->s_feature_incompat &
+	    EXT3_FEATURE_INCOMPAT_EXTENTS)
+		inode.i_flags |= EXT4_EXTENTS_FL;
 	if (debugfs_write_new_inode(newfile, &inode, argv[0])) {
 		close(fd);
 		return;
@@ -1469,7 +1695,7 @@ void do_mknod(int argc, char *argv[])
 	ext2fs_mark_ib_dirty(current_fs);
 	memset(&inode, 0, sizeof(inode));
 	inode.i_mode = mode;
-	inode.i_atime = inode.i_ctime = inode.i_mtime = 
+	inode.i_atime = inode.i_ctime = inode.i_mtime =
 		current_fs->now ? current_fs->now : time(0);
 	if ((major < 256) && (minor < 256)) {
 		inode.i_block[0] = major*256+minor;
@@ -1513,7 +1739,7 @@ try_again:
 	if (retval == EXT2_ET_DIR_NO_SPACE) {
 		retval = ext2fs_expand_dir(current_fs, parent);
 		if (retval) {
-			com_err("argv[0]", retval, "while expanding directory");
+			com_err(argv[0], retval, "while expanding directory");
 			return;
 		}
 		goto try_again;
@@ -1526,7 +1752,7 @@ try_again:
 }
 
 static int release_blocks_proc(ext2_filsys fs, blk_t *blocknr,
-			       int blockcnt EXT2FS_ATTR((unused)), 
+			       int blockcnt EXT2FS_ATTR((unused)),
 			       void *private EXT2FS_ATTR((unused)))
 {
 	blk_t	block;
@@ -1626,7 +1852,7 @@ static int rmdir_proc(ext2_ino_t dir EXT2FS_ATTR((unused)),
 	rds->empty = 0;
 	return 0;
 }
-	
+
 void do_rmdir(int argc, char *argv[])
 {
 	int retval;
@@ -1683,7 +1909,7 @@ void do_rmdir(int argc, char *argv[])
 	}
 }
 
-void do_show_debugfs_params(int argc EXT2FS_ATTR((unused)), 
+void do_show_debugfs_params(int argc EXT2FS_ATTR((unused)),
 			    char *argv[] EXT2FS_ATTR((unused)))
 {
 	FILE *out = stdout;
@@ -1712,7 +1938,7 @@ void do_expand_dir(int argc, char *argv[])
 void do_features(int argc, char *argv[])
 {
 	int	i;
-	
+
 	if (check_fs_open(argv[0]))
 		return;
 
@@ -1735,7 +1961,7 @@ void do_bmap(int argc, char *argv[])
 	blk_t		blk, pblk;
 	int		err;
 	errcode_t	errcode;
-	
+
 	if (common_args_process(argc, argv, 3, 3, argv[0],
 				"<file> logical_blk", 0))
 		return;
@@ -1775,7 +2001,7 @@ void do_imap(int argc, char *argv[])
 			group);
 		return;
 	}
-	block_nr = current_fs->group_desc[(unsigned)group].bg_inode_table + 
+	block_nr = current_fs->group_desc[(unsigned)group].bg_inode_table +
 		block;
 	offset &= (EXT2_BLOCK_SIZE(current_fs->super) - 1);
 
@@ -1805,6 +2031,71 @@ void do_set_current_time(int argc, char *argv[])
 	}
 }
 
+static int find_supp_feature(__u32 *supp, int feature_type, char *name)
+{
+	int compat, bit, ret;
+	unsigned int feature_mask;
+
+	if (name) {
+		if (feature_type == E2P_FS_FEATURE)
+			ret = e2p_string2feature(name, &compat, &feature_mask);
+		else
+			ret = e2p_jrnl_string2feature(name, &compat,
+						      &feature_mask);
+		if (ret)
+			return ret;
+
+		if (!(supp[compat] & feature_mask))
+			return 1;
+	} else {
+	        for (compat = 0; compat < 3; compat++) {
+		        for (bit = 0, feature_mask = 1; bit < 32;
+			     bit++, feature_mask <<= 1) {
+			        if (supp[compat] & feature_mask) {
+					if (feature_type == E2P_FS_FEATURE)
+						fprintf(stdout, " %s",
+						e2p_feature2string(compat,
+						feature_mask));
+					else
+						fprintf(stdout, " %s",
+						e2p_jrnl_feature2string(compat,
+						feature_mask));
+				}
+	        	}
+		}
+	        fprintf(stdout, "\n");
+	}
+
+	return 0;
+}
+
+void do_supported_features(int argc, char *argv[])
+{
+        int	ret;
+	__u32	supp[3] = { EXT2_LIB_FEATURE_COMPAT_SUPP,
+			    EXT2_LIB_FEATURE_INCOMPAT_SUPP,
+			    EXT2_LIB_FEATURE_RO_COMPAT_SUPP };
+	__u32	jrnl_supp[3] = { JFS_KNOWN_COMPAT_FEATURES,
+				 JFS_KNOWN_INCOMPAT_FEATURES,
+				 JFS_KNOWN_ROCOMPAT_FEATURES };
+
+	if (argc > 1) {
+		ret = find_supp_feature(supp, E2P_FS_FEATURE, argv[1]);
+		if (ret) {
+			ret = find_supp_feature(jrnl_supp, E2P_JOURNAL_FEATURE,
+						argv[1]);
+		}
+		if (ret)
+			com_err(argv[0], 0, "Unknown feature: %s\n", argv[1]);
+		else
+			fprintf(stdout, "Supported feature: %s\n", argv[1]);
+	} else {
+		fprintf(stdout, "Supported features:");
+		ret = find_supp_feature(supp, E2P_FS_FEATURE, NULL);
+		ret = find_supp_feature(jrnl_supp, E2P_JOURNAL_FEATURE, NULL);
+	}
+}
+
 static int source_file(const char *cmd_file, int sci_idx)
 {
 	FILE		*f;
@@ -1822,6 +2113,8 @@ static int source_file(const char *cmd_file, int sci_idx)
 			exit(1);
 		}
 	}
+	fflush(stdout);
+	fflush(stderr);
 	setbuf(stdout, NULL);
 	setbuf(stderr, NULL);
 	while (!feof(f)) {
@@ -1840,6 +2133,8 @@ static int source_file(const char *cmd_file, int sci_idx)
 			exit_status++;
 		}
 	}
+	if (f != stdin)
+		fclose(f);
 	return exit_status;
 }
 
@@ -1857,7 +2152,7 @@ int main(int argc, char **argv)
 	blk_t		blocksize = 0;
 	int		catastrophic = 0;
 	char		*data_filename = 0;
-	
+
 	if (debug_prog_name == 0)
 		debug_prog_name = "debugfs";
 
@@ -1865,7 +2160,7 @@ int main(int argc, char **argv)
 	fprintf (stderr, "%s %s (%s)\n", debug_prog_name,
 		 E2FSPROGS_VERSION, E2FSPROGS_DATE);
 
-	while ((c = getopt (argc, argv, "iwcR:f:b:s:Vd:")) != EOF) {
+	while ((c = getopt (argc, argv, "iwcR:f:b:s:Vd:D")) != EOF) {
 		switch (c) {
 		case 'R':
 			request = optarg;
@@ -1882,12 +2177,15 @@ int main(int argc, char **argv)
 		case 'w':
 			open_flags |= EXT2_FLAG_RW;
 			break;
+		case 'D':
+			open_flags |= EXT2_FLAG_DIRECT_IO;
+			break;
 		case 'b':
-			blocksize = parse_ulong(optarg, argv[0], 
+			blocksize = parse_ulong(optarg, argv[0],
 						"block size", 0);
 			break;
 		case 's':
-			superblock = parse_ulong(optarg, argv[0], 
+			superblock = parse_ulong(optarg, argv[0],
 						 "superblock number", 0);
 			break;
 		case 'c':
@@ -1907,7 +2205,7 @@ int main(int argc, char **argv)
 		open_filesystem(argv[optind], open_flags,
 				superblock, blocksize, catastrophic,
 				data_filename);
-	
+
 	sci_idx = ss_create_invocation(debug_prog_name, "0.0", (char *) NULL,
 				       &debug_cmds, &retval);
 	if (retval) {
@@ -1940,9 +2238,11 @@ int main(int argc, char **argv)
 		ss_listen(sci_idx);
 	}
 
+	ss_delete_invocation(sci_idx);
+
 	if (current_fs)
 		close_filesystem();
-	
+
 	remove_error_table(&et_ext2_error_table);
 	return exit_status;
 }
