@@ -1,6 +1,6 @@
 /*
  * htree.c --- hash tree routines
- * 
+ *
  * Copyright (C) 2002 Theodore Ts'o.  This file may be redistributed
  * under the terms of the GNU Public License.
  */
@@ -17,12 +17,14 @@
 #include <sys/types.h>
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#else 
+#else
 extern int optind;
 extern char *optarg;
 #endif
 
 #include "debugfs.h"
+#include "uuid/uuid.h"
+#include "e2p/e2p.h"
 
 static FILE *pager;
 
@@ -38,9 +40,10 @@ static void htree_dump_leaf_node(ext2_filsys fs, ext2_ino_t ino,
 	char		name[EXT2_NAME_LEN + 1];
 	char		tmp[EXT2_NAME_LEN + 16];
 	blk_t		pblk;
-	ext2_dirhash_t 	hash;
+	ext2_dirhash_t 	hash, minor_hash;
+	unsigned int	rec_len;
 	int		hash_alg;
-	
+
 	errcode = ext2fs_bmap(fs, ino, inode, buf, 0, blk, &pblk);
 	if (errcode) {
 		com_err("htree_dump_leaf_node", errcode,
@@ -48,10 +51,13 @@ static void htree_dump_leaf_node(ext2_filsys fs, ext2_ino_t ino,
 		return;
 	}
 
+	printf("Reading directory block %lu, phys %lu\n",
+	       (unsigned long) blk, (unsigned long) pblk);
 	errcode = ext2fs_read_dir_block2(current_fs, pblk, buf, 0);
 	if (errcode) {
 		com_err("htree_dump_leaf_node", errcode,
-			"while 	reading block %u\n", blk);
+			"while reading block %lu (%lu)\n",
+			(unsigned long) blk, (unsigned long) pblk);
 		return;
 	}
 	hash_alg = rootnode->hash_version;
@@ -61,10 +67,17 @@ static void htree_dump_leaf_node(ext2_filsys fs, ext2_ino_t ino,
 
 	while (offset < fs->blocksize) {
 		dirent = (struct ext2_dir_entry *) (buf + offset);
-		if (((offset + dirent->rec_len) > fs->blocksize) ||
-		    (dirent->rec_len < 8) ||
-		    ((dirent->rec_len % 4) != 0) ||
-		    (((dirent->name_len & 0xFF)+8) > dirent->rec_len)) {
+		errcode = ext2fs_get_rec_len(fs, dirent, &rec_len);
+		if (errcode) {
+			com_err("htree_dump_leaf_inode", errcode,
+				"while getting rec_len for block %lu",
+				(unsigned long) blk);
+			return;
+		}
+		if (((offset + rec_len) > fs->blocksize) ||
+		    (rec_len < 8) ||
+		    ((rec_len % 4) != 0) ||
+		    ((((unsigned) dirent->name_len & 0xFF)+8) > rec_len)) {
 			fprintf(pager, "Corrupted directory block (%u)!\n", blk);
 			break;
 		}
@@ -74,12 +87,12 @@ static void htree_dump_leaf_node(ext2_filsys fs, ext2_ino_t ino,
 		name[thislen] = '\0';
 		errcode = ext2fs_dirhash(hash_alg, name,
 					 thislen, fs->super->s_hash_seed,
-					 &hash, 0);
+					 &hash, &minor_hash);
 		if (errcode)
 			com_err("htree_dump_leaf_node", errcode,
 				"while calculating hash");
-		sprintf(tmp, "%u 0x%08x (%d) %s   ", dirent->inode,
-			hash, dirent->rec_len, name);
+		sprintf(tmp, "%u 0x%08x-%08x (%d) %s   ", dirent->inode,
+			hash, minor_hash, rec_len, name);
 		thislen = strlen(tmp);
 		if (col + thislen > 80) {
 			fprintf(pager, "\n");
@@ -87,7 +100,7 @@ static void htree_dump_leaf_node(ext2_filsys fs, ext2_ino_t ino,
 		}
 		fprintf(pager, "%s", tmp);
 		col += thislen;
-		offset += dirent->rec_len;
+		offset += rec_len;
 	}
 	fprintf(pager, "\n");
 }
@@ -102,13 +115,13 @@ static void htree_dump_int_block(ext2_filsys fs, ext2_ino_t ino,
 static void htree_dump_int_node(ext2_filsys fs, ext2_ino_t ino,
 				struct ext2_inode *inode,
 				struct ext2_dx_root_info * rootnode,
-				struct ext2_dx_entry *ent, 
+				struct ext2_dx_entry *ent,
 				char *buf, int level)
 {
 	struct ext2_dx_countlimit	limit;
 	struct ext2_dx_entry		e;
 	int				hash, i;
-	
+
 
 	limit = *((struct ext2_dx_countlimit *) ent);
 	limit.count = ext2fs_le16_to_cpu(limit.count);
@@ -156,7 +169,7 @@ static void htree_dump_int_block(ext2_filsys fs, ext2_ino_t ino,
 		fprintf(pager, "Couldn't allocate child block.\n");
 		return;
 	}
-	
+
 	errcode = ext2fs_bmap(fs, ino, inode, buf, 0, blk, &pblk);
 	if (errcode) {
 		com_err("htree_dump_int_block", errcode,
@@ -186,6 +199,7 @@ void do_htree_dump(int argc, char *argv[])
 	struct ext2_inode inode;
 	int		c;
 	int		long_opt = 0;
+	blk_t		blk;
 	char		*buf = NULL;
 	struct 		ext2_dx_root_info  *rootnode;
 	struct 		ext2_dx_entry *ent;
@@ -228,7 +242,7 @@ void do_htree_dump(int argc, char *argv[])
 		com_err(argv[0], 0, "Not a directory");
 		goto errout;
 	}
-	
+
 	if ((inode.i_flags & EXT2_BTREE_FL) == 0) {
 		com_err(argv[0], 0, "Not a hash-indexed directory");
 		goto errout;
@@ -240,7 +254,14 @@ void do_htree_dump(int argc, char *argv[])
 		goto errout;
 	}
 
-	errcode = io_channel_read_blk(current_fs->io, inode.i_block[0], 
+	errcode = ext2fs_bmap(current_fs, ino, &inode, buf, 0, 0, &blk);
+	if (errcode) {
+		com_err("do_htree_block", errcode,
+			"while mapping logical block 0\n");
+		goto errout;
+	}
+
+	errcode = io_channel_read_blk(current_fs->io, blk,
 				      1, buf);
 	if (errcode) {
 		com_err(argv[0], errcode, "Error reading root node");
@@ -264,8 +285,7 @@ void do_htree_dump(int argc, char *argv[])
 			    rootnode->indirect_levels);
 
 errout:
-	if (buf)
-		free(buf);
+	free(buf);
 	close_pager(pager);
 }
 
@@ -279,14 +299,23 @@ void do_dx_hash(int argc, char *argv[])
 	int		c;
 	int		hash_version = 0;
 	__u32		hash_seed[4];
-	
+
 	hash_seed[0] = hash_seed[1] = hash_seed[2] = hash_seed[3] = 0;
 
 	reset_getopt();
-	while ((c = getopt (argc, argv, "h:")) != EOF) {
+	while ((c = getopt (argc, argv, "h:s:")) != EOF) {
 		switch (c) {
 		case 'h':
-			hash_version = atoi(optarg);
+			hash_version = e2p_string2hash(optarg);
+			if (hash_version < 0)
+				hash_version = atoi(optarg);
+			break;
+		case 's':
+			if (uuid_parse(optarg, (unsigned char *) hash_seed)) {
+				fprintf(stderr, "Invalid UUID format: %s\n",
+					optarg);
+				return;
+			}
 			break;
 		default:
 			goto print_usage;
@@ -294,7 +323,8 @@ void do_dx_hash(int argc, char *argv[])
 	}
 	if (optind != argc-1) {
 	print_usage:
-		com_err(argv[0], 0, "usage: dx_hash filename");
+		com_err(argv[0], 0, "usage: dx_hash [-h hash_alg] "
+			"[-s hash_seed] filename");
 		return;
 	}
 	err = ext2fs_dirhash(hash_version, argv[optind], strlen(argv[optind]),
@@ -319,14 +349,14 @@ struct process_block_struct {
 };
 
 static int search_dir_block(ext2_filsys fs, blk_t *blocknr,
-			    e2_blkcnt_t blockcnt, blk_t ref_blk, 
+			    e2_blkcnt_t blockcnt, blk_t ref_blk,
 			    int ref_offset, void *priv_data);
 
 void do_dirsearch(int argc, char *argv[])
 {
 	ext2_ino_t	inode;
 	struct process_block_struct pb;
-	
+
 	if (check_fs_open(argv[0]))
 		return;
 
@@ -346,7 +376,7 @@ void do_dirsearch(int argc, char *argv[])
 	}
 	pb.search_name = argv[2];
 	pb.len = strlen(pb.search_name);
-	
+
 	ext2fs_block_iterate2(current_fs, inode, BLOCK_FLAG_READ_ONLY, 0,
 			      search_dir_block, &pb);
 
@@ -355,7 +385,7 @@ void do_dirsearch(int argc, char *argv[])
 
 
 static int search_dir_block(ext2_filsys fs, blk_t *blocknr,
-			    e2_blkcnt_t blockcnt, 
+			    e2_blkcnt_t blockcnt,
 			    blk_t ref_blk EXT2FS_ATTR((unused)),
 			    int ref_offset EXT2FS_ATTR((unused)),
 			    void *priv_data)
@@ -364,6 +394,7 @@ static int search_dir_block(ext2_filsys fs, blk_t *blocknr,
 	struct ext2_dir_entry *dirent;
 	errcode_t	       	errcode;
 	unsigned int		offset = 0;
+	unsigned int		rec_len;
 
 	if (blockcnt < 0)
 		return 0;
@@ -379,9 +410,15 @@ static int search_dir_block(ext2_filsys fs, blk_t *blocknr,
 
 	while (offset < fs->blocksize) {
 		dirent = (struct ext2_dir_entry *) (p->buf + offset);
-
+		errcode = ext2fs_get_rec_len(fs, dirent, &rec_len);
+		if (errcode) {
+			com_err("htree_dump_leaf_inode", errcode,
+				"while getting rec_len for block %lu",
+				(unsigned long) *blocknr);
+			return BLOCK_ABORT;
+		}
 		if (dirent->inode &&
-		    p->len == (dirent->name_len & 0xFF) && 
+		    p->len == (dirent->name_len & 0xFF) &&
 		    strncmp(p->search_name, dirent->name,
 			    p->len) == 0) {
 			printf("Entry found at logical block %lld, "
@@ -390,7 +427,7 @@ static int search_dir_block(ext2_filsys fs, blk_t *blocknr,
 			printf("offset %u\n", offset);
 			return BLOCK_ABORT;
 		}
-		offset += dirent->rec_len;
+		offset += rec_len;
 	}
 	return 0;
 }
