@@ -6,9 +6,11 @@
  *                                 Universite Pierre et Marie Curie (Paris VI)
  *
  * Copyright (C) 1995, 1996, 1997  Theodore Ts'o <tytso@mit.edu>
- * 
- * This file can be redistributed under the terms of the GNU Library General
- * Public License
+ *
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
+ * %End-Header%
  */
 
 #include <stdio.h>
@@ -209,6 +211,8 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	print_features(sb, f);
 	print_super_flags(sb, f);
 	print_mntopts(sb, f);
+	if (sb->s_mount_opts[0])
+		fprintf(f, "Mount options:            %s\n", sb->s_mount_opts);
 	fprintf(f, "Filesystem state:        ");
 	print_fs_state (f, sb->s_state);
 	fprintf(f, "\n");
@@ -227,7 +231,7 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	fprintf(f, "Block size:               %u\n", EXT2_BLOCK_SIZE(sb));
 	fprintf(f, "Fragment size:            %u\n", EXT2_FRAG_SIZE(sb));
 	if (sb->s_reserved_gdt_blocks)
-		fprintf(f, "Reserved GDT blocks:      %u\n", 
+		fprintf(f, "Reserved GDT blocks:      %u\n",
 			sb->s_reserved_gdt_blocks);
 	fprintf(f, "Blocks per group:         %u\n", sb->s_blocks_per_group);
 	fprintf(f, "Fragments per group:      %u\n", sb->s_frags_per_group);
@@ -267,6 +271,24 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 		next = sb->s_lastcheck + sb->s_checkinterval;
 		fprintf(f, "Next check after:         %s", ctime(&next));
 	}
+#define POW2(x) ((__u64) 1 << (x))
+	if (sb->s_kbytes_written) {
+		fprintf(f, "Lifetime writes:          ");
+		if (sb->s_kbytes_written < POW2(13))
+			fprintf(f, "%llu kB\n", sb->s_kbytes_written);
+		else if (sb->s_kbytes_written < POW2(23))
+			fprintf(f, "%llu MB\n",
+				(sb->s_kbytes_written + POW2(9)) >> 10);
+		else if (sb->s_kbytes_written < POW2(33))
+			fprintf(f, "%llu GB\n",
+				(sb->s_kbytes_written + POW2(19)) >> 20);
+		else if (sb->s_kbytes_written < POW2(43))
+			fprintf(f, "%llu TB\n",
+				(sb->s_kbytes_written + POW2(29)) >> 30);
+		else
+			fprintf(f, "%llu PB\n",
+				(sb->s_kbytes_written + POW2(39)) >> 40);
+	}
 	fprintf(f, "Reserved blocks uid:      ");
 	print_user(sb->s_def_resuid, f);
 	fprintf(f, "Reserved blocks gid:      ");
@@ -275,10 +297,10 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 		fprintf(f, "First inode:              %d\n", sb->s_first_ino);
 		fprintf(f, "Inode size:	          %d\n", sb->s_inode_size);
 		if (sb->s_min_extra_isize)
-			fprintf(f, "Required extra isize:     %d\n", 
+			fprintf(f, "Required extra isize:     %d\n",
 				sb->s_min_extra_isize);
 		if (sb->s_want_extra_isize)
-			fprintf(f, "Desired extra isize:      %d\n", 
+			fprintf(f, "Desired extra isize:      %d\n",
 				sb->s_want_extra_isize);
 	}
 	if (!e2p_is_null_uuid(sb->s_journal_uuid))
@@ -310,6 +332,48 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 			fprintf(f, "type %u\n", sb->s_jnl_backup_type);
 		}
 	}
+	if (sb->s_snapshot_inum) {
+		fprintf(f, "Snapshot inode:           %u\n",
+			sb->s_snapshot_inum);
+		fprintf(f, "Snapshot ID:              %u\n",
+			sb->s_snapshot_id);
+		fprintf(f, "Snapshot reserved blocks: %llu\n",
+			sb->s_snapshot_r_blocks_count);
+	}
+	if (sb->s_snapshot_list)
+		fprintf(f, "Snapshot list head:       %u\n",
+			sb->s_snapshot_list);
+	if (sb->s_error_count)
+		fprintf(f, "FS Error count:           %u\n",
+			sb->s_error_count);
+	if (sb->s_first_error_time) {
+		tm = sb->s_first_error_time;
+		fprintf(f, "First error time:         %s", ctime(&tm));
+		memset(buf, 0, sizeof(buf));
+		strncpy(buf, sb->s_first_error_func,
+			sizeof(sb->s_first_error_func));
+		fprintf(f, "First error function:     %s\n", buf);
+		fprintf(f, "First error line #:       %u\n",
+			sb->s_first_error_line);
+		fprintf(f, "First error inode #:      %u\n",
+			sb->s_first_error_ino);
+		fprintf(f, "First error block #:      %llu\n",
+			sb->s_first_error_block);
+	}
+	if (sb->s_last_error_time) {
+		tm = sb->s_last_error_time;
+		fprintf(f, "Last error time:          %s", ctime(&tm));
+		memset(buf, 0, sizeof(buf));
+		strncpy(buf, sb->s_last_error_func,
+			sizeof(sb->s_last_error_func));
+		fprintf(f, "Last error function:      %s\n", buf);
+		fprintf(f, "Last error line #:        %u\n",
+			sb->s_last_error_line);
+		fprintf(f, "Last error inode #:       %u\n",
+			sb->s_last_error_ino);
+		fprintf(f, "Last error block #:       %llu\n",
+			sb->s_last_error_block);
+	}
 }
 
 void list_super (struct ext2_super_block * s)
