@@ -5,8 +5,10 @@
  *                                 Laboratoire MASI, Institut Blaise Pascal
  *                                 Universite Pierre et Marie Curie (Paris VI)
  *
- * This file can be redistributed under the terms of the GNU Library General
- * Public License
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
+ * %End-Header%
  */
 
 /*
@@ -16,8 +18,6 @@
 
 #include <stdio.h>
 
-#include <linux/ext2_fs.h>
-
 #include "e2p.h"
 
 void print_fs_errors (FILE * f, unsigned short errors)
