@@ -2,12 +2,11 @@
  * dblist_dir.c --- iterate by directory entry
  *
  * Copyright 1997 by Theodore Ts'o
- * 
+ *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
  * %End-Header%
- * 
  */
 
 #include <stdio.h>
@@ -17,12 +16,7 @@
 #include <string.h>
 #include <time.h>
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fsP.h"
 
 static int db_dir_proc(ext2_filsys fs, struct ext2_db_entry *db_info,
@@ -31,7 +25,7 @@ static int db_dir_proc(ext2_filsys fs, struct ext2_db_entry *db_info,
 errcode_t ext2fs_dblist_dir_iterate(ext2_dblist dblist,
 				    int	flags,
 				    char	*block_buf,
-				    int (*func)(ino_t	dir,
+				    int (*func)(ext2_ino_t dir,
 						int	entry,
 						struct ext2_dir_entry *dirent,
 						int	offset,
@@ -50,20 +44,18 @@ errcode_t ext2fs_dblist_dir_iterate(ext2_dblist dblist,
 	if (block_buf)
 		ctx.buf = block_buf;
 	else {
-		retval = ext2fs_get_mem(dblist->fs->blocksize,
-					(void **) &ctx.buf);
+		retval = ext2fs_get_mem(dblist->fs->blocksize, &ctx.buf);
 		if (retval)
 			return retval;
 	}
-	ctx.func = 0;
-	ctx.func2 = func;
+	ctx.func = func;
 	ctx.priv_data = priv_data;
 	ctx.errcode = 0;
 
 	retval = ext2fs_dblist_iterate(dblist, db_dir_proc, &ctx);
-	
+
 	if (!block_buf)
-		ext2fs_free_mem((void **) &ctx.buf);
+		ext2fs_free_mem(&ctx.buf);
 	if (retval)
 		return retval;
 	return ctx.errcode;
@@ -73,10 +65,15 @@ static int db_dir_proc(ext2_filsys fs, struct ext2_db_entry *db_info,
 		       void *priv_data)
 {
 	struct dir_context	*ctx;
+	int			ret;
 
 	ctx = (struct dir_context *) priv_data;
 	ctx->dir = db_info->ino;
-	
-	return ext2fs_process_dir_block(fs, &db_info->blk,
-					db_info->blockcnt, 0, 0, priv_data);
+	ctx->errcode = 0;
+
+	ret = ext2fs_process_dir_block(fs, &db_info->blk,
+				       db_info->blockcnt, 0, 0, priv_data);
+	if ((ret & BLOCK_ABORT) && !ctx->errcode)
+		return DBLIST_ABORT;
+	return 0;
 }
