@@ -2,10 +2,13 @@
  * dosio.c -- Disk I/O module for the ext2fs/DOS library.
  *
  * Copyright (c) 1997 by Theodore Ts'o.
- * 
- * Copyright (c) 1997 Mark Habersack
- * This file may be distributed under the terms of the GNU Public License.
  *
+ * Copyright (c) 1997 Mark Habersack
+ *
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
+ * %End-Header%
  */
 
 #include <stdio.h>
@@ -179,7 +182,7 @@ static errcode_t dos_open(const char *dev, int flags, io_channel *channel)
   PARTITION      *part;
   PTABLE_ENTRY   *pent;
   PARTITION        **newparts;
-  
+
   if(!dev)
   {
     _dio_error = ERR_BADDEV;
@@ -278,8 +281,8 @@ static errcode_t dos_open(const char *dev, int flags, io_channel *channel)
   if(!HW_OK())
   {
     _dio_error = ERR_HARDWARE;
-    if (part)
-	    free(part);
+    free(part->dev);
+    free(part);
     return EFAULT;
   }
 
@@ -298,8 +301,8 @@ static errcode_t dos_open(const char *dev, int flags, io_channel *channel)
   if(!HW_OK())
   {
     _dio_error = ERR_HARDWARE;
-    if (part)
-	    free(part);
+    free(part->dev);
+    free(part);
     return EFAULT;
   }
 
@@ -310,8 +313,8 @@ static errcode_t dos_open(const char *dev, int flags, io_channel *channel)
   {
     _dio_error = part->pno == 0xFE ? ERR_EMPTYPART :
                  part->pno == 0xFD ? ERR_LINUXSWAP : ERR_NOTEXT2FS;
-    if (part)
-	    free(part);
+    free(part->dev);
+    free(part);
     return ENODEV;
   }
 
@@ -352,10 +355,8 @@ static errcode_t dos_open(const char *dev, int flags, io_channel *channel)
 
 static errcode_t dos_close(io_channel channel)
 {
-	if (channel->name)
-		free(channel->name);
-	if (channel)
-		free(channel);
+	free(channel->name);
+	free(channel);
 
 	return 0;
 }
