@@ -1,12 +1,15 @@
 /*
  * get_pathname.c --- do directry/inode -> name translation
- * 
+ *
  * Copyright (C) 1993, 1994, 1995 Theodore Ts'o.
  *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
  * %End-Header%
+ */
+
+/*
  *
  * 	ext2fs_get_pathname(fs, dir, ino, name)
  *
@@ -15,7 +18,7 @@
  * 	directory inode, and <ino> is the inode number itself.  If
  * 	<ino> is zero, then ext2fs_get_pathname will return pathname
  * 	of the the directory <dir>.
- * 
+ *
  */
 
 #include <stdio.h>
@@ -24,28 +27,23 @@
 #include <unistd.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 struct get_pathname_struct {
-	ino_t		search_ino;
-	ino_t		parent;
+	ext2_ino_t	search_ino;
+	ext2_ino_t	parent;
 	char		*name;
 	errcode_t	errcode;
 };
 
 #ifdef __TURBOC__
-#pragma argsused
+ #pragma argsused
 #endif
 static int get_pathname_proc(struct ext2_dir_entry *dirent,
-			     int	offset,
-			     int	blocksize,
-			     char	*buf,
+			     int	offset EXT2FS_ATTR((unused)),
+			     int	blocksize EXT2FS_ATTR((unused)),
+			     char	*buf EXT2FS_ATTR((unused)),
 			     void	*priv_data)
 {
 	struct get_pathname_struct	*gp;
@@ -58,7 +56,7 @@ static int get_pathname_proc(struct ext2_dir_entry *dirent,
 		gp->parent = dirent->inode;
 	if (dirent->inode == gp->search_ino) {
 		retval = ext2fs_get_mem((dirent->name_len & 0xFF) + 1,
-					(void **) &gp->name);
+					&gp->name);
 		if (retval) {
 			gp->errcode = retval;
 			return DIRENT_ABORT;
@@ -70,15 +68,16 @@ static int get_pathname_proc(struct ext2_dir_entry *dirent,
 	return 0;
 }
 
-static errcode_t ext2fs_get_pathname_int(ext2_filsys fs, ino_t dir, ino_t ino,
-					 int maxdepth, char *buf, char **name)
+static errcode_t ext2fs_get_pathname_int(ext2_filsys fs, ext2_ino_t dir,
+					 ext2_ino_t ino, int maxdepth,
+					 char *buf, char **name)
 {
 	struct get_pathname_struct gp;
 	char	*parent_name, *ret;
 	errcode_t	retval;
 
 	if (dir == ino) {
-		retval = ext2fs_get_mem(2, (void **)name);
+		retval = ext2fs_get_mem(2, name);
 		if (retval)
 			return retval;
 		strcpy(*name, (dir == EXT2_ROOT_INO) ? "/" : ".");
@@ -86,7 +85,7 @@ static errcode_t ext2fs_get_pathname_int(ext2_filsys fs, ino_t dir, ino_t ino,
 	}
 
 	if (!dir || (maxdepth < 0)) {
-		retval = ext2fs_get_mem(4, (void **)name);
+		retval = ext2fs_get_mem(4, name);
 		if (retval)
 			return retval;
 		strcpy(*name, "...");
@@ -97,7 +96,7 @@ static errcode_t ext2fs_get_pathname_int(ext2_filsys fs, ino_t dir, ino_t ino,
 	gp.parent = 0;
 	gp.name = 0;
 	gp.errcode = 0;
-	
+
 	retval = ext2fs_dir_iterate(fs, dir, 0, buf, get_pathname_proc, &gp);
 	if (retval)
 		goto cleanup;
@@ -114,16 +113,15 @@ static errcode_t ext2fs_get_pathname_int(ext2_filsys fs, ino_t dir, ino_t ino,
 		*name = parent_name;
 		return 0;
 	}
-	
-	if (gp.name) 
+
+	if (gp.name)
 		retval = ext2fs_get_mem(strlen(parent_name)+strlen(gp.name)+2,
-					(void **) &ret);
+					&ret);
 	else
-		retval = ext2fs_get_mem(strlen(parent_name)+5,
-					(void **) &ret);
+		retval = ext2fs_get_mem(strlen(parent_name)+5, &ret);
 	if (retval)
 		goto cleanup;
-	
+
 	ret[0] = 0;
 	if (parent_name[1])
 		strcat(ret, parent_name);
@@ -133,16 +131,16 @@ static errcode_t ext2fs_get_pathname_int(ext2_filsys fs, ino_t dir, ino_t ino,
 	else
 		strcat(ret, "???");
 	*name = ret;
-	ext2fs_free_mem((void **) &parent_name);
+	ext2fs_free_mem(&parent_name);
 	retval = 0;
-	
+
 cleanup:
 	if (gp.name)
-		ext2fs_free_mem((void **) &gp.name);
+		ext2fs_free_mem(&gp.name);
 	return retval;
 }
 
-errcode_t ext2fs_get_pathname(ext2_filsys fs, ino_t dir, ino_t ino,
+errcode_t ext2fs_get_pathname(ext2_filsys fs, ext2_ino_t dir, ext2_ino_t ino,
 			      char **name)
 {
 	char	*buf;
@@ -150,13 +148,13 @@ errcode_t ext2fs_get_pathname(ext2_filsys fs, ino_t dir, ino_t ino,
 
 	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
 
-	retval = ext2fs_get_mem(fs->blocksize, (void **) &buf);
+	retval = ext2fs_get_mem(fs->blocksize, &buf);
 	if (retval)
 		return retval;
 	if (dir == ino)
 		ino = 0;
 	retval = ext2fs_get_pathname_int(fs, dir, ino, 32, buf, name);
-	ext2fs_free_mem((void **) &buf);
+	ext2fs_free_mem(&buf);
 	return retval;
-	
+
 }
