@@ -4,8 +4,8 @@
  * Copyright (C) 1994, 1995, 1996, 1997 Theodore Ts'o.
  *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
  * %End-Header%
  */
 
@@ -46,11 +46,11 @@ extern long long llseek (int fd, long long offset, int origin);
 
 #else	/* ! HAVE_LLSEEK */
 
-#if defined(__alpha__) || defined (__ia64__)
+#if SIZEOF_LONG == SIZEOF_LONG_LONG
 
 #define llseek lseek
 
-#else /* !__alpha__ && !__ia64__*/
+#else /* SIZEOF_LONG != SIZEOF_LONG_LONG */
 
 #include <linux/unistd.h>
 
@@ -74,7 +74,7 @@ static ext2_loff_t my_llseek (int fd, ext2_loff_t offset, int origin)
 
 #ifndef __i386__
 	retval = _llseek(fd, ((unsigned long long) offset) >> 32,
-#else			  
+#else
 	retval = syscall(__NR__llseek, fd, (unsigned long long) (offset >> 32),
 #endif
 			  ((unsigned long long) offset) & 0xffffffff,
@@ -100,7 +100,7 @@ ext2_loff_t ext2fs_llseek (int fd, ext2_loff_t offset, int origin)
 		errno = EINVAL;
 		return -1;
 	}
-	
+
 	result = my_llseek (fd, offset, origin);
 	if (result == -1 && errno == ENOSYS) {
 		/*
@@ -121,12 +121,16 @@ ext2_loff_t ext2fs_llseek (int fd, ext2_loff_t offset, int origin)
 
 ext2_loff_t ext2fs_llseek (int fd, ext2_loff_t offset, int origin)
 {
+#if defined(HAVE_LSEEK64) && defined(HAVE_LSEEK64_PROTOTYPE)
+	return lseek64 (fd, offset, origin);
+#else
 	if ((sizeof(off_t) < sizeof(ext2_loff_t)) &&
 	    (offset >= ((ext2_loff_t) 1 << ((sizeof(off_t)*8) -1)))) {
 		errno = EINVAL;
 		return -1;
 	}
 	return lseek (fd, (off_t) offset, origin);
+#endif
 }
 
 #endif 	/* linux */
