@@ -4,8 +4,8 @@
  * Copyright (C) 1997 Theodore Ts'o.
  *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
  * %End-Header%
  */
 
@@ -27,11 +27,17 @@ static const char *lib_date = E2FSPROGS_DATE;
 int ext2fs_parse_version_string(const char *ver_string)
 {
 	const char *cp;
-	int version = 0;
+	int version = 0, dot_count = 0;
 
 	for (cp = ver_string; *cp; cp++) {
+		if (*cp == '.') {
+			if (dot_count++)
+				break;
+			else
+				continue;
+		}
 		if (!isdigit(*cp))
-			continue;
+			break;
 		version = (version * 10) + (*cp - '0');
 	}
 	return version;
