@@ -14,9 +14,19 @@
 #ifdef HAS_STDLIB_H
 #include <stdlib.h>
 #endif
+#ifdef HAVE_ERRNO_H
+#include <errno.h>
+#else
+extern int errno;
+#endif
 #include "ss_internal.h"
 #include <stdio.h>
 
+static int check_request_table PROTOTYPE((ss_request_table *rqtbl, int argc,
+					  char *argv[], int sci_idx));
+static int really_execute_command PROTOTYPE((int sci_idx, int argc,
+					     char **argv[]));
+
 /*
  * get_request(tbl, idx)
  *
@@ -196,7 +206,7 @@ int ss_execute_line (sci_idx, line_ptr)
     char *line_ptr;
 {
     char **argv;
-    int argc;
+    int argc, ret;
 
     /* flush leading whitespace */
     while (line_ptr[0] == ' ' || line_ptr[0] == '\t')
@@ -208,16 +218,21 @@ int ss_execute_line (sci_idx, line_ptr)
             return SS_ET_ESCAPE_DISABLED;
         else {
             line_ptr++;
-            system(line_ptr);
-	    return 0;
+            return (system(line_ptr) < 0) ? errno : 0;
         }
     }
 
     /* parse it */
     argv = ss_parse(sci_idx, line_ptr, &argc);
-    if (argc == 0)
+    if (argc == 0) {
+	free(argv);
         return 0;
+    }
 
     /* look it up in the request tables, execute if found */
-    return really_execute_command (sci_idx, argc, &argv);
+    ret = really_execute_command (sci_idx, argc, &argv);
+
+    free(argv);
+
+    return(ret);
 }
